"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import io
import json
import os
import shutil
from datetime import datetime
from unittest import mock

from nose.tools import assert_equal, assert_is_instance, assert_true
from panther_core.data_model import _DATAMODEL_FOLDER
from pyfakefs.fake_filesystem_unittest import TestCase, Pause
from schema import SchemaWrongKeyError

from panther_analysis_tool import main as pat
from panther_analysis_tool import util
from panther_analysis_tool.backend.client import BackendError
from panther_analysis_tool.backend.mocks import MockBackend

FIXTURES_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../', 'fixtures'))
DETECTIONS_FIXTURES_PATH = os.path.join(FIXTURES_PATH, 'detections')

print('Using fixtures path:', FIXTURES_PATH)


def _mock_invoke(**_kwargs):  # pylint: disable=C0103
    return {
        "Payload": io.BytesIO(json.dumps({
            "statusCode": 400,
            "body": "another upload is in process",
        }).encode("utf-8")),
        "StatusCode": 400,
    }


class TestPantherAnalysisTool(TestCase):
    def setUp(self):
        # Data Models and Globals write the source code to a file and import it as module.
        # This will not work if we are simply writing on the in-memory, fake filesystem.
        # We thus copy to a temporary space the Data Model Python modules.
        self.data_model_modules = [os.path.join(DETECTIONS_FIXTURES_PATH,
                                                'valid_analysis/data_models/GSuite.Events.DataModel.py')]
        for data_model_module in self.data_model_modules:
            shutil.copy(data_model_module, _DATAMODEL_FOLDER)
        os.makedirs(pat.TMP_HELPER_MODULE_LOCATION, exist_ok=True)
        self.global_modules = {
            "panther": os.path.join(DETECTIONS_FIXTURES_PATH, 'valid_analysis/global_helpers/helpers.py'),
            "a_helper": os.path.join(DETECTIONS_FIXTURES_PATH, 'valid_analysis/global_helpers/a_helper.py'),
            "b_helper": os.path.join(DETECTIONS_FIXTURES_PATH, 'valid_analysis/global_helpers/b_helper.py'),
        }
        for module_name, filename in self.global_modules.items():
            shutil.copy(filename, os.path.join(pat.TMP_HELPER_MODULE_LOCATION, f"{module_name}.py"))
        self.setUpPyfakefs()
        self.fs.add_real_directory(FIXTURES_PATH)
        self.fs.add_real_directory(pat.TMP_HELPER_MODULE_LOCATION, read_only=False)

    def tearDown(self) -> None:
        with Pause(self.fs):
            for data_model_module in self.data_model_modules:
                os.remove(os.path.join(_DATAMODEL_FOLDER, os.path.split(data_model_module)[-1]))

    def test_valid_json_policy_spec(self):
        for spec_filename, _, loaded_spec, _ in pat.load_analysis_specs([DETECTIONS_FIXTURES_PATH], ignore_files=[]):
            if spec_filename.endswith('example_policy.json'):
                assert_is_instance(loaded_spec, dict)
                assert_true(loaded_spec != {})

    def test_ignored_files_are_not_loaded(self):
        for spec_filename, _, loaded_spec, _ in pat.load_analysis_specs([DETECTIONS_FIXTURES_PATH],
                                                                        ignore_files=["./example_ignored.yml"]):
            assert_true(loaded_spec != "example_ignored.yml")

    def test_multiple_ignored_files_are_not_loaded(self):
        for spec_filename, _, loaded_spec, _ in pat.load_analysis_specs([DETECTIONS_FIXTURES_PATH],
                                                                        ignore_files=["./example_ignored.yml",
                                                                                      "./example_ignored_multi.yml"]):
            assert_true(loaded_spec != "example_ignored.yml" and loaded_spec != "example_ignored_multi.yml")

    def test_sdk_file_is_read_and_overrides_cli(self):
        for spec_filename, _, loaded_spec, _ in pat.load_analysis_specs([DETECTIONS_FIXTURES_PATH], ignore_files=[]):
            assert_true(loaded_spec != "example_ignored.yml" and loaded_spec != "example_ignored_multi.yml")

    def test_valid_yaml_policy_spec(self):
        for spec_filename, _, loaded_spec, _ in pat.load_analysis_specs([DETECTIONS_FIXTURES_PATH], ignore_files=[]):
            if spec_filename.endswith('example_policy.yml'):
                assert_is_instance(loaded_spec, dict)
                assert_true(loaded_spec != {})

    def test_valid_pack_spec(self):
        pack_loaded = False
        for spec_filename, _, loaded_spec, _ in pat.load_analysis_specs([DETECTIONS_FIXTURES_PATH], ignore_files=[]):
            if spec_filename.endswith('sample-pack.yml'):
                assert_is_instance(loaded_spec, dict)
                assert_true(loaded_spec != {})
                pack_loaded = True
        assert_true(pack_loaded)

    def test_datetime_converted(self):
        test_date = datetime.now()
        test_date_string = pat.datetime_converted(test_date)
        assert_is_instance(test_date_string, str)

    def test_handle_wrong_key_error(self):
        sample_keys = ['DisplayName', 'Enabled', 'Filename']
        expected_output = '{} not in list of valid keys: {}'
        # test successful regex match and correct error returned
        test_str = "Wrong key 'DisplaName' in {'DisplaName':'one','Enabled':true, 'Filename':'sample'}"
        exc = SchemaWrongKeyError(test_str)
        err = pat.handle_wrong_key_error(exc, sample_keys)
        assert_equal(str(err), expected_output.format("'DisplaName'", sample_keys))
        # test failing regex match
        test_str = "Will not match"
        exc = SchemaWrongKeyError(test_str)
        err = pat.handle_wrong_key_error(exc, sample_keys)
        assert_equal(str(err), expected_output.format("UNKNOWN_KEY", sample_keys))

    def test_load_policy_specs_from_folder(self):
        args = pat.setup_parser().parse_args(f'test --path {DETECTIONS_FIXTURES_PATH}'.split())
        args.filter_inverted = {}
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)
        assert_equal(invalid_specs[0][0],
                     f'{DETECTIONS_FIXTURES_PATH}/example_malformed_policy.yml')
        assert_equal(len(invalid_specs), 12)

    def test_policies_from_folder(self):
        args = pat.setup_parser().parse_args(f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis/policies'.split())
        args.filter_inverted = {}
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_rules_from_folder(self):
        args = pat.setup_parser().parse_args(f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis/rules'.split())
        args.filter_inverted = {}
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_queries_from_folder(self):
        args = pat.setup_parser().parse_args(f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis/queries'.split())
        args.filter_inverted = {}
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_scheduled_rules_from_folder(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis/scheduled_rules'.split())
        args.filter_inverted = {}
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_rules_from_current_dir(self):
        # This is a work around to test running tool against current directory
        return_code = -1
        invalid_specs = None
        valid_rule_path = os.path.join(DETECTIONS_FIXTURES_PATH, 'valid_analysis/policies')
        # test default path, '.'
        with Pause(self.fs):
            original_path = os.getcwd()
            try:
                os.chdir(valid_rule_path)
                args = pat.setup_parser().parse_args('test'.split())
                args.filter_inverted = {}
                return_code, invalid_specs = pat.test_analysis(args)
            finally:
                os.chdir(original_path)
        # asserts are outside of the pause to ensure the fakefs gets resumed
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)
        return_code = -1
        invalid_specs = None
        # test explicitly setting current dir
        with Pause(self.fs):
            original_path = os.getcwd()
            os.chdir(valid_rule_path)
            args = pat.setup_parser().parse_args('test --path ./'.split())
            args.filter_inverted = {}
            return_code, invalid_specs = pat.test_analysis(args)
            os.chdir(original_path)
        # asserts are outside of the pause to ensure the fakefs gets resumed
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_parse_filters(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --filter AnalysisType=policy,global Severity=Critical Enabled=true'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        assert_true('AnalysisType' in args.filter.keys())
        assert_true('policy' in args.filter['AnalysisType'])
        assert_true('global' in args.filter['AnalysisType'])
        assert_true('Severity' in args.filter.keys())
        assert_true('Critical' in args.filter['Severity'])
        assert_true('Enabled' in args.filter.keys())
        assert_true(True in args.filter['Enabled'])

    def test_with_filters(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --filter AnalysisType=policy,global'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_enabled_filter(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/disabled_rule --filter Enabled=true'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_enabled_filter_inverted(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/disabled_rule --filter Enabled!=false'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_aws_profiles(self):
        aws_profile = 'AWS_PROFILE'
        args = pat.setup_parser().parse_args(
            f'upload --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --aws-profile myprofile'.split())
        util.set_env(aws_profile, args.aws_profile)
        assert_equal('myprofile', args.aws_profile)
        assert_equal(args.aws_profile, os.environ.get(aws_profile))

    def test_invalid_rule_definition(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH} --filter RuleID=AWS.CloudTrail.MFAEnabled'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)
        assert_equal(len(invalid_specs), 8)

    def test_invalid_rule_test(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH} --filter RuleID=Example.Rule.Invalid.Test'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)
        assert_equal(len(invalid_specs), 8)

    def test_invalid_characters(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH} --filter Severity=High ResourceTypes=AWS.IAM.User'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)
        assert_equal(len(invalid_specs), 9)

    def test_unknown_exception(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH} --filter RuleID=Example.Rule.Unknown.Exception'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)
        assert_equal(len(invalid_specs), 8)

    def test_with_invalid_mocks(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH} --filter Severity=Critical RuleID=Example.Rule.Invalid.Mock'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)
        assert_equal(len(invalid_specs), 8)

    def test_with_tag_filters(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --filter Tags=AWS,CIS'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_with_tag_filters_inverted(self):
        # Note: a comparison of the tests passed is required to make this test robust
        # (8 passing vs 1 passing)
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --filter Tags=AWS,CIS Tags!=SOC2'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_with_minimum_tests(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --minimum-tests 1'.split())
        args.filter_inverted = {}
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)
        assert_equal(len(invalid_specs), 0)

    def test_with_minimum_tests_failing(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --minimum-tests 2'.split())
        args.filter_inverted = {}
        return_code, invalid_specs = pat.test_analysis(args)
        # Failing, because some of the fixtures only have one test case
        assert_equal(return_code, 1)
        assert_equal(len(invalid_specs), 0)

    def test_with_minimum_tests_no_passing(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH} --filter PolicyID=IAM.MFAEnabled.Required.Tests --minimum-tests 2'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        # Failing, because while there are two unit tests they both have expected result False
        assert_equal(return_code, 1)
        assert_equal(len(invalid_specs), 8)

    def test_invalid_resource_type(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH} --filter PolicyID=Example.Bad.Resource.Type'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)
        assert_equal(len(invalid_specs), 8)

    def test_invalid_log_type(self):
        args = pat.setup_parser().parse_args(
            f'test --path {DETECTIONS_FIXTURES_PATH} --filter RuleID=Example.Bad.Log.Type'.split())
        args.filter, args.filter_inverted = pat.parse_filter(args.filter)
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)
        self.equal = assert_equal(len(invalid_specs), 8)

    def test_zip_analysis(self):
        # Note: This is a workaround for CI
        try:
            self.fs.create_dir('tmp/')
        except OSError:
            pass
        args = pat.setup_parser(). \
            parse_args(f'zip --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --out tmp/'.split())

        return_code, out_filename = pat.zip_analysis(args)
        assert_true(out_filename.startswith("tmp/"))
        statinfo = os.stat(out_filename)
        assert_true(statinfo.st_size > 0)
        assert_equal(return_code, 0)
        assert_true(out_filename.endswith('.zip'))

    def test_zip_analysis_chunks(self):
        # Note: This is a workaround for CI
        try:
            self.fs.create_dir('tmp/')
        except OSError:
            pass
        args = pat.setup_parser(). \
            parse_args(f'upload --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --out tmp/ --batch'.split())

        results = pat.zip_analysis_chunks(args)
        for out_filename in results:
            assert_true(out_filename.startswith("tmp/"))
            statinfo = os.stat(out_filename)
            assert_true(statinfo.st_size > 0)
            assert_true(out_filename.endswith('.zip'))

        assert_equal(6, len(results))

    def test_generate_release_assets(self):
        # Note: This is a workaround for CI
        try:
            self.fs.create_dir('tmp/release/')
        except OSError:
            pass

        args = pat.setup_parser().parse_args(
            f'release --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --out tmp/release/'.split())
        return_code, _ = pat.generate_release_assets(args)
        analysis_file = 'tmp/release/panther-analysis-all.zip'
        statinfo = os.stat(analysis_file)
        assert_true(statinfo.st_size > 0)
        assert_equal(return_code, 0)

    def test_retry_uploads(self):
        import logging

        backend = MockBackend()
        backend.bulk_upload = mock.MagicMock(
            side_effect=BackendError("another upload is in process")
        )

        args = pat.setup_parser().parse_args(
            f'--debug upload --path {DETECTIONS_FIXTURES_PATH}/valid_analysis'.split())

        # fails max of 10 times on default
        with mock.patch('time.sleep', return_value=None) as time_mock:
            with mock.patch.multiple(logging, debug=mock.DEFAULT, warning=mock.DEFAULT,
                                     info=mock.DEFAULT) as logging_mocks:
                return_code, _ = pat.upload_analysis(backend, args)
                assert_equal(return_code, 1)
                assert_equal(logging_mocks['debug'].call_count, 21)
                assert_equal(logging_mocks['warning'].call_count, 1)
                # test + zip + upload messages
                assert_equal(logging_mocks['info'].call_count, 3)
                assert_equal(time_mock.call_count, 10)

        # invalid retry count, default to 0
        args = pat.setup_parser().parse_args(
            f'--debug upload --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --max-retries -1'.split())
        with mock.patch('time.sleep', return_value=None) as time_mock:
            with mock.patch.multiple(logging, debug=mock.DEFAULT, warning=mock.DEFAULT,
                                     info=mock.DEFAULT) as logging_mocks:
                return_code, _ = pat.upload_analysis(backend, args)
                assert_equal(return_code, 1)
                assert_equal(logging_mocks['debug'].call_count, 1)
                assert_equal(logging_mocks['warning'].call_count, 2)
                assert_equal(logging_mocks['info'].call_count, 3)
                assert_equal(time_mock.call_count, 0)

        # invalid retry count, default to 10
        args = pat.setup_parser().parse_args(
            f'--debug upload --path {DETECTIONS_FIXTURES_PATH}/valid_analysis --max-retries 100'.split())
        with mock.patch('time.sleep', return_value=None) as time_mock:
            with mock.patch.multiple(logging, debug=mock.DEFAULT, warning=mock.DEFAULT,
                                     info=mock.DEFAULT) as logging_mocks:
                return_code, _ = pat.upload_analysis(backend, args)
                assert_equal(return_code, 1)
                assert_equal(logging_mocks['debug'].call_count, 21)
                # warning about max and final error
                assert_equal(logging_mocks['warning'].call_count, 2)
                assert_equal(logging_mocks['info'].call_count, 3)
                assert_equal(time_mock.call_count, 10)

    def test_available_destination_names_invalid_name_returned(self):
        """When an available destination is given but does not match the returned names"""
        args = pat.setup_parser().parse_args(f'test --path {DETECTIONS_FIXTURES_PATH}/valid_analysis '
                                             '--available-destination Pagerduty'.split())
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 1)

    def test_available_destination_names_valid_name_returned(self):
        """When an available destination is given but matches the returned name"""
        args = pat.setup_parser().parse_args(f'test '
                                             f'--path '
                                             f' {DETECTIONS_FIXTURES_PATH}/destinations '
                                             '--available-destination Pagerduty'.split())
        return_code, invalid_specs = pat.test_analysis(args)
        assert_equal(return_code, 0)



import unittest
from pathlib import Path

import panther_analysis_tool


class Version(unittest.TestCase):
    def test_version_matches(self) -> None:
        root = Path(__file__).parent.parent.parent.parent
        v1 = (root / "VERSION").read_text().strip()
        v2 = panther_analysis_tool.constants.VERSION_STRING.split(" ")[1]

        self.assertEqual(v1, v2)



from typing import Any

from panther_analysis_tool.backend.client import (
    BackendCheckResponse,
    BackendResponse,
    BulkUploadParams,
    BulkUploadResponse,
)
from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import (
    DeleteDetectionsParams,
    DeleteSavedQueriesParams,
    ListSchemasParams,
    PantherSDKBulkUploadParams,
    UpdateManagedSchemaParams,
)


class MockBackend(BackendClient):
    def bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:
        pass

    def check(self) -> BackendCheckResponse:
        pass

    def list_managed_schemas(self, params: ListSchemasParams) -> BackendResponse[Any]:
        pass

    def update_managed_schema(self, params: UpdateManagedSchemaParams) -> BackendResponse[Any]:
        pass

    def delete_saved_queries(self, params: DeleteSavedQueriesParams) -> BackendResponse[Any]:
        pass

    def delete_detections(self, params: DeleteDetectionsParams) -> BackendResponse[Any]:
        pass

    def panthersdk_bulk_upload(self, params: PantherSDKBulkUploadParams) -> BackendResponse[Any]:
        pass



from pathlib import Path
from setuptools import setup, find_packages

this_directory = Path(__file__).parent
PAT_VERSION = (this_directory / "VERSION").read_text().strip()

install_requires = [
    'gql',
    'aiohttp',
    'boto3',
    'dynaconf',
    'jsonpath-ng',
    'requests',
    'ruamel.yaml',
    'schema',
    'semver',
    'panther_core',
    'typing-extensions',
    'jsonlines',
]

with open('requirements.txt') as f:
    dependencies_with_versions = []
    for dependency in f.readlines():
        dependency_with_version = dependency.strip()
        package_name = dependency_with_version.split('==')[0]
        if package_name in install_requires:
            dependencies_with_versions.append(dependency_with_version)

setup(
    name='panther_analysis_tool',
    version=PAT_VERSION,
    packages=find_packages(),
    license='AGPL-3.0',
    description=
    'Panther command line interface for writing, testing, and packaging policies/rules.',
    author='Panther Labs Inc',
    author_email='pypi@runpanther.io',
    url='https://github.com/panther-labs/panther_analysis_tool',
    download_url=f'https://github.com/panther-labs/panther_analysis_tool/archive/v{PAT_VERSION}.tar.gz',
    keywords=['Security', 'CLI'],
    scripts=['bin/panther_analysis_tool'],
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Security',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 3.9',
    ],
    include_package_data=True,
)



import logging
import os
import runpy
import sys
from typing import Dict, List

from jsonlines import jsonlines

from panther_analysis_tool.panthersdk.models import (
    DataModel,
    Detection,
    SdkContent,
    SdkContentType,
    panther_sdk_key_to_type,
)


def run_sdk_module(panther_sdk_cache_path: str) -> None:
    if not os.path.exists(os.path.join("panther_content", "__main__.py")):
        raise FileNotFoundError("Did not find a Panther SDK based module at ./panther_content")

    try:
        os.remove(panther_sdk_cache_path)
    except FileNotFoundError:
        pass

    path_had_cwd = os.getcwd() in sys.path
    if not path_had_cwd:
        sys.path.append(os.getcwd())
    runpy.run_module("panther_content")
    if not path_had_cwd:
        sys.path.remove(os.getcwd())

    if not os.path.exists(panther_sdk_cache_path):
        logging.error("panther_content did not generate %s", panther_sdk_cache_path)
        raise FileNotFoundError(f"panther_content did not generate {panther_sdk_cache_path}")


def get_sdk_cache_path() -> str:
    return os.path.join(".panther", "panther-sdk-cache")


def load_intermediate_sdk_cache(panther_sdk_cache_path: str) -> List[Dict]:
    """Load intermediate Panther SDK code from the cache.

    Returns:
        A list of the intermediate json as dicts.
    """
    if not os.path.exists(panther_sdk_cache_path):
        raise FileNotFoundError(f"No file exists with path {panther_sdk_cache_path}")

    with jsonlines.open(panther_sdk_cache_path) as reader:
        return [obj for obj in reader]  # pylint: disable=R1721


def unmarshal_sdk_intermediates(intermediates: List[Dict]) -> SdkContent:
    sdk_content = SdkContent()

    for intermediate in intermediates:
        sdk_type = panther_sdk_key_to_type(intermediate.get("key") or "")
        if sdk_type is SdkContentType.RULE:
            sdk_content.detections.append(Detection(intermediate))
        if sdk_type is SdkContentType.SCHEDULED_RULE:
            sdk_content.detections.append(Detection(intermediate))
        if sdk_type is SdkContentType.POLICY:
            sdk_content.detections.append(Detection(intermediate))
        if sdk_type is SdkContentType.DATA_MODEL:
            sdk_content.data_models.append(DataModel(intermediate))

    return sdk_content



import argparse
import logging
from typing import Final, Tuple

from panther_analysis_tool import panthersdk
from panther_analysis_tool.backend.client import BackendError
from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import PantherSDKBulkUploadParams


def run(
    backend: BackendClient,
    args: argparse.Namespace,  # pylint: disable=unused-argument
    indirect_invocation: bool = False,
) -> Tuple[int, str]:
    """Packages and uploads all policies and rules from the Panther SDK-based module at
    ./panther_content, if it exists, into a Panther deployment.

        Returns 1 if the packaging or upload fails.

        Args:
            backend: Backend API client.
            args: The populated Argparse namespace with parsed command-line arguments.
            indirect_invocation: True if this function is being invoked as part of
                                 another command (probably the legacy bulk upload command)

        Returns:
            Return code
    """

    panther_sdk_cache_path: Final = panthersdk.get_sdk_cache_path()

    try:
        panthersdk.run_sdk_module(panther_sdk_cache_path)
    except FileNotFoundError:
        err_message = "Did not find a Panther SDK based module at ./panther_content"
        if indirect_invocation:
            # If this is run automatically at the end of the standard upload command,
            # this isn't an error that should cause the invocation to return 1.
            logging.debug(err_message)
            return 0, ""
        logging.error(err_message)
        return 1, ""

    with open(panther_sdk_cache_path) as sdk_cache_file:
        try:
            result = backend.panthersdk_bulk_upload(
                params=PantherSDKBulkUploadParams(content=sdk_cache_file.read())
            )
        except BackendError as exc:
            logging.error(exc)
            return 1, ""

    logging.info("Panther SDK module upload succeeded")
    logging.info(
        "(Rules: %d new; %d modified; %d total)",
        result.data.rules.new,
        result.data.rules.modified,
        result.data.rules.total,
    )
    logging.info(
        "(Policies: %d new; %d modified; %d total)",
        result.data.policies.new,
        result.data.policies.modified,
        result.data.policies.total,
    )
    logging.info(
        "(Queries: %d new; %d modified; %d total)",
        result.data.queries.new,
        result.data.queries.modified,
        result.data.queries.total,
    )
    logging.info(
        "(Data Models: %d new; %d modified; %d total)",
        result.data.data_models.new,
        result.data.data_models.modified,
        result.data.data_models.total,
    )

    return 0, ""



"""
 Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import logging
from typing import List, Tuple

from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import (
    DeleteDetectionsParams,
    DeleteSavedQueriesParams,
)


def run(backend: BackendClient, args: argparse.Namespace) -> Tuple[int, str]:
    # pylint: disable=too-many-return-statements

    logging.info("preparing bulk delete...")

    # Get lists of detection ids and query names from args
    query_name_list = args.query_id
    detection_id_list = args.analysis_id

    targets_detections = len(detection_id_list) != 0
    targets_saved_queries = len(query_name_list) != 0

    if not targets_detections and not targets_saved_queries:
        logging.error("Must specify a list of analysis or queries to delete")
        logging.error("Run panther_analysis_tool -h for help statement")
        return 1, ""

    # Dry Run: Detections
    if targets_detections:
        code, msg = _delete_detections_dry_run(backend, detection_id_list)
        if code != 0:
            return code, msg

    # Dry Run: Saved Queries
    if targets_saved_queries:
        code, msg = _delete_queries_dry_run(backend, query_name_list)
        if code != 0:
            return code, msg

    # Prompt for user confirmation (unless bypassed)
    if not args.confirm_bypass:
        confirm = input("\nContinue? (y/n) ")

        if confirm.lower() != "y":
            print("Cancelled")
            return 0, ""

        print("")

    # Delete Detections
    if targets_detections:
        code, msg = _delete_detections(backend, detection_id_list)
        if code != 0:
            logging.warning("error deleting detections: %s", msg)
            return code, msg

        logging.info("successfully deleted detections.")

    # Delete Saved Queries
    if targets_saved_queries:
        code, msg = _delete_queries(backend, query_name_list)
        if code != 0:
            logging.warning("error deleting saved queries: %s", msg)
            return code, msg

        logging.info("successfully deleted saved queries.")

    logging.info("done")
    return 0, ""


def _delete_detections_dry_run(backend: BackendClient, ids: List[str]) -> Tuple[int, str]:
    if len(ids) == 0:
        return 0, ""

    res = backend.delete_detections(
        DeleteDetectionsParams(dry_run=True, ids=ids, include_saved_queries=True)
    )

    if res.status_code != 200:
        logging.error("Error connecting to backend.")
        return 1, ""

    found_ids = res.data.ids or []

    for detection_id in ids:
        if detection_id in found_ids:
            logging.info("Detection '%s' will be deleted", detection_id)
        else:
            logging.info("Detection '%s' was not found.", detection_id)

    linked_query_names = res.data.saved_query_names or []

    for query_id in linked_query_names:
        logging.info("Linked saved query '%s' will be deleted.", query_id)

    return 0, ""


def _delete_queries_dry_run(backend: BackendClient, names: List[str]) -> Tuple[int, str]:
    if len(names) == 0:
        return 0, ""

    res = backend.delete_saved_queries(
        DeleteSavedQueriesParams(dry_run=True, names=names, include_detections=True)
    )

    if res.status_code != 200:
        logging.error("Error connecting to backend.")
        return 1, ""

    found_names = res.data.names or []

    for query_name in names:
        if query_name in found_names:
            logging.info("Saved query '%s' will be deleted", query_name)
        else:
            logging.info("Saved query '%s' was not found.", query_name)

    linked_detection_ids = res.data.detection_ids or []

    for detection_id in linked_detection_ids:
        logging.info("Linked detection '%s' will be deleted.", detection_id)

    return 0, ""


def _delete_detections(backend: BackendClient, ids: List[str]) -> Tuple[int, str]:
    if len(ids) == 0:
        return 0, ""

    delete_detections_res = backend.delete_detections(
        DeleteDetectionsParams(
            ids=ids,
            dry_run=False,
            include_saved_queries=True,
        )
    )

    if delete_detections_res.status_code != 200:
        logging.error("Error deleting detections")
        return 1, ""

    logging.info(
        "%d detections and %d linked saved queries deleted",
        len(delete_detections_res.data.ids),
        len(delete_detections_res.data.saved_query_names),
    )

    return 0, ""


def _delete_queries(backend: BackendClient, names: List[str]) -> Tuple[int, str]:
    delete_queries_res = backend.delete_saved_queries(
        DeleteSavedQueriesParams(
            names=names,
            dry_run=False,
            include_detections=True,
        )
    )

    if delete_queries_res.status_code != 200:
        logging.error("Error deleting saved queries")
        return 1, ""

    logging.info(
        "%d saved queries and %d linked detections deleted",
        len(delete_queries_res.data.names),
        len(delete_queries_res.data.detection_ids),
    )

    return 0, ""



import os
import tempfile
from typing import Dict, Final

from schema import Schema

from panther_analysis_tool.schemas import (
    DATA_MODEL_SCHEMA,
    GLOBAL_SCHEMA,
    LOOKUP_TABLE_SCHEMA,
    PACK_SCHEMA,
    POLICY_SCHEMA,
    RULE_SCHEMA,
    SCHEDULED_QUERY_SCHEMA,
)

VERSION_STRING: Final = "panther_analysis_tool 0.19.12"

CONFIG_FILE = ".panther_settings.yml"
DATA_MODEL_LOCATION = "./data_models"
HELPERS_LOCATION = "./global_helpers"
LUTS_LOCATION = "./lookup_tables"
DATA_MODEL_PATH_PATTERN = "*data_models*"
LUTS_PATH_PATTERN = "*lookup_tables*"
HELPERS_PATH_PATTERN = "*/global_helpers"
PACKS_PATH_PATTERN = "*/packs"
POLICIES_PATH_PATTERN = "*policies*"
QUERIES_PATH_PATTERN = "*queries*"
RULES_PATH_PATTERN = "*rules*"
TMP_HELPER_MODULE_LOCATION = os.path.join(tempfile.gettempdir(), "panther-path", "globals")

DATAMODEL = "datamodel"
DETECTION = "detection"
GLOBAL = "global"
LOOKUP_TABLE = "lookup_table"
PACK = "pack"
POLICY = "policy"
QUERY = "scheduled_query"
RULE = "rule"
SCHEDULED_RULE = "scheduled_rule"

RESERVED_FUNCTIONS = (
    "alert_context",
    "dedup",
    "description",
    "destinations",
    "reference",
    "runbook",
    "severity",
    "title",
)

VALID_SEVERITIES = ["INFO", "LOW", "MEDIUM", "HIGH", "CRITICAL"]

SCHEMAS: Dict[str, Schema] = {
    DATAMODEL: DATA_MODEL_SCHEMA,
    GLOBAL: GLOBAL_SCHEMA,
    LOOKUP_TABLE: LOOKUP_TABLE_SCHEMA,
    PACK: PACK_SCHEMA,
    POLICY: POLICY_SCHEMA,
    QUERY: SCHEDULED_QUERY_SCHEMA,
    RULE: RULE_SCHEMA,
    SCHEDULED_RULE: RULE_SCHEMA,
}

SET_FIELDS = [
    "LogTypes",
    "PackIDs",
    "OutputIds",
    "SummaryAttributes",
    "Suppressions",
    "Tags",
]



import argparse
import logging
from collections import defaultdict
from typing import Dict, Final, List, Optional, Tuple

import panther_core.rule

from panther_analysis_tool import panthersdk


class TestSummary:
    def __init__(self) -> None:
        self.failed_tests: Dict[str, List[panthersdk.UnitTest]] = defaultdict(
            list
        )  # detection id to list of unit tests
        self.fail_count = 0
        self.pass_count = 0

    def total_count(self) -> int:
        return self.pass_count + self.fail_count

    def add_failure(self, detection_id: str, test: panthersdk.UnitTest) -> None:
        self.failed_tests[detection_id].append(test)
        self.fail_count += 1

    def test_passed(self) -> None:
        self.pass_count += 1

    def summary(self) -> str:
        summary = "--------------------------\nPanther CLI Test Summary\n"
        summary += f"    Passed: {self.pass_count}\n"
        summary += f"    Failed: {self.fail_count}\n"
        summary += f"    Total:  {self.total_count()}\n\n"

        if len(self.failed_tests) > 0:
            summary += "--------------------------\nFailed Tests Summary\n"

            for detection_id, failures in self.failed_tests.items():
                summary += f"   {detection_id}\n"
                for unit_test in failures:
                    summary += f"       {unit_test.name} ({unit_test.origin})\n"
                    for reason in unit_test.fail_reasons:
                        summary += f"            {reason}\n"
                    summary += "\n"  # new line between tests
        return summary

    def tests_failed(self) -> bool:
        return self.fail_count > 0


_TEST_SUMMARY = TestSummary()


def run(args: argparse.Namespace, indirect_invocation: bool = False) -> Tuple[int, list]:
    """Runs unit tests for Panther SDK detections.

    Args:
        args: The populated Argparse namespace with parsed command-line arguments.
        indirect_invocation: True if this function is being invoked as part of
                                 another command (probably the legacy test command)

    Returns:
        A tuple of the return code, and a list of tuples containing invalid specs and their error.
    """
    panther_sdk_cache_path: Final = panthersdk.get_sdk_cache_path()

    try:
        panthersdk.run_sdk_module(panther_sdk_cache_path)
        intermediates = panthersdk.load_intermediate_sdk_cache(panther_sdk_cache_path)
        sdk_content = panthersdk.unmarshal_sdk_intermediates(intermediates)
        detections = _filter_detections(args, sdk_content.detections)
        logging.info("Running Unit Tests for Panther Content\n")
        tests_failed = _run_unit_tests(detections, sdk_content.data_models, args.minimum_tests)
        return int(tests_failed), []
    except FileNotFoundError as err:
        if indirect_invocation:
            # If this is run automatically at the end of the standard test command,
            # this isn't an error that should cause the invocation to return 1.
            logging.debug(err)
            return 0, []
        logging.error(err)
        return 1, []


def _filter_detections(
    args: argparse.Namespace, detections: List[panthersdk.Detection]
) -> List[panthersdk.Detection]:
    """Filters out the detections to be tested by using the command line args.

    Args:
        args: The populated Argparse namespace with parsed command-line arguments.
        detections: The list of detections that need to be filtered for testing.

    Returns:
        A list of detections to be unit tested.
    """
    filtered = []
    for detection in detections:
        # filter out detections with no unit tests
        if not detection.has_unit_tests():
            continue

        # filter out using filter arg TODO

        # filter using enabled only arg
        if args.skip_disabled_tests and detection.disabled():
            continue

        filtered.append(detection)

    return filtered


def _run_unit_tests(
    detections: List[panthersdk.Detection],
    data_models: List[panthersdk.DataModel],
    min_tests: int = 0,
) -> bool:
    """Runs the unit tests for the given detections, printing out test results and a summary.

    Args:
        detections: The list of detections to be unit tested.
        min_tests: The minimum number of unit tests each detection needs to have.

    Returns:
        True if at least one unit test failed, False if all tests pass.
    """

    # organize data models by log_type
    data_models_by_log_type = {}
    for data_model in data_models:
        if data_model.log_type in data_models_by_log_type:
            logging.error(
                "Conflicting Data Model (%s) for LogType %s",
                data_model.data_model_id,
                data_model.log_type,
            )

            return True  # data model conflicting is considered failing
        data_models_by_log_type[data_model.log_type] = data_model.to_panther_core_data_model()

    for detection in detections:
        print(detection.detection_id)
        if len(detection.filters) == 0:
            print("    Detection had no filters to test")
            continue

        for unit_test in detection.unit_tests:
            logs_data_model: Optional[panthersdk.DataModel] = data_models_by_log_type.get(
                unit_test.data.get("p_log_type")  # get the log type from the test event
            )
            detection_result = detection.to_panther_core_detection().run(
                panther_core.PantherEvent(unit_test.data, logs_data_model),  # pylint: disable=E1101
                outputs={},
                outputs_names={},
            )
            result = detection_result.detection_output

            if detection_result.detection_exception:
                unit_test.add_fail_reason(str(detection_result.detection_exception))
                print(
                    f"    [FAIL] {unit_test.name}: An exception occured while running the unit test: "
                    f"{str(detection_result.detection_exception)}"
                )
                _TEST_SUMMARY.add_failure(detection.detection_id, unit_test)
            elif result != unit_test.expect_match:
                reason = f"Expected match to be {unit_test.expect_match} but got {result}"
                unit_test.add_fail_reason(reason)
                print(f"    [FAIL] {unit_test.name}: {reason}")
                _TEST_SUMMARY.add_failure(detection.detection_id, unit_test)
            else:
                print(f"    [PASS] {unit_test.name}")
                _TEST_SUMMARY.test_passed()

        has_pass_and_fail_tests = False
        if min_tests >= 2:
            has_pass_and_fail_tests = detection.has_pass_and_fail_tests()

        n_tests = len(detection.unit_tests)
        if n_tests < min_tests or (min_tests >= 2 and not has_pass_and_fail_tests):
            # create a fake unit test to represent the minimum tests failure
            unit_test = panthersdk.UnitTest({})
            unit_test.name = "minimum required tests"
            unit_test.origin = detection.origin  # origin will be the origin of the detection
            if n_tests < min_tests:
                unit_test.add_fail_reason(
                    f"Insufficient test coverage, {min_tests} tests required but only {n_tests} found"
                )
            if min_tests >= 2 and not has_pass_and_fail_tests:
                unit_test.add_fail_reason(
                    "Insufficient test coverage: expected at least one passing and one failing test"
                )
            _TEST_SUMMARY.add_failure(detection.detection_id, unit_test)

        print()  # print blank line in between tests

    print(_TEST_SUMMARY.summary())
    return _TEST_SUMMARY.tests_failed()



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy
import json
from unittest import TestCase

from panther_analysis_tool.immutable import ImmutableCaseInsensitiveDict, ImmutableList, json_encoder


class TestImmutableDict(TestCase):

    def setUp(self) -> None:
        self.initial_dict = {'t': 10, 'a': [{'b': 1, 'c': 2}], 'd': {'e': {'f': True}}}
        self.immutable_dict = ImmutableCaseInsensitiveDict(self.initial_dict)

    def test_assignment_not_allowed(self) -> None:
        with self.assertRaises(TypeError):
            # pylint: disable=E1137
            self.immutable_dict['a'] = 1  # type: ignore

    def test_nested_assignment_not_allowed(self) -> None:
        with self.assertRaises(TypeError):
            # pylint: disable=E1137
            self.immutable_dict['d']['e']['f'] = False

    def test_original_dict_not_mutated(self) -> None:
        _ = self.immutable_dict['a']
        self.assertEqual(self.initial_dict, self.immutable_dict._container)

    def test_raises_error_for_non_existent_key(self) -> None:
        with self.assertRaises(KeyError):
            _ = self.immutable_dict['a-non-existent-key']

    def test_getitem(self) -> None:
        self.assertEqual(self.immutable_dict['t'], self.initial_dict['t'])

    def test_nested_access(self) -> None:
        self.assertEqual(self.immutable_dict['a'][0]['b'], 1)
        self.assertEqual(self.immutable_dict['d']['e']['f'], True)

    def test_equality(self) -> None:
        # Equality with dict
        self.assertEqual(self.immutable_dict, self.initial_dict)
        # Equality with another instance
        equal_dict = deepcopy(self.initial_dict)
        other_immutable_dict = ImmutableCaseInsensitiveDict(equal_dict)
        self.assertEqual(other_immutable_dict, self.immutable_dict)

    def test_shallow_copy(self) -> None:
        self.assertEqual(self.immutable_dict._container, self.initial_dict)
        self.assertIsNot(self.immutable_dict._container, self.initial_dict)

    def test_get(self) -> None:
        self.assertIsInstance(self.immutable_dict.get('d'), ImmutableCaseInsensitiveDict)
        self.assertIsInstance(self.immutable_dict.get('a'), ImmutableList)

    def test_ensure_immutable(self) -> None:
        initial_dict = {'a': [[1, 2], [3, 4]], 'b': {'c': {'d': 1}}, 't': 10, 'e': {'f': [{'g': 90}]}}
        immutable_dict = ImmutableCaseInsensitiveDict(initial_dict)
        # List of lists with immutable elements
        self.assertIsInstance(immutable_dict['a'], ImmutableList)
        self.assertIsInstance(immutable_dict['a'][0], ImmutableList)
        self.assertEqual(immutable_dict['a'][0][1], 2)
        # Two-level nested dictionary
        self.assertIsInstance(immutable_dict['b'], ImmutableCaseInsensitiveDict)
        self.assertIsInstance(immutable_dict['b']['c'], ImmutableCaseInsensitiveDict)
        self.assertEqual(immutable_dict['b']['c']['d'], 1)
        # Plain immutable object at top-level
        self.assertIsInstance(immutable_dict['t'], int)
        self.assertEqual(immutable_dict['t'], 10)
        # Two-level dictionary with nested list as value
        self.assertIsInstance(immutable_dict['e']['f'], ImmutableList)
        self.assertIsInstance(immutable_dict['e']['f'][0], ImmutableCaseInsensitiveDict)
        self.assertEqual(immutable_dict['e']['f'][0]['g'], 90)

    def test_copy(self) -> None:
        initial_dict = {'a': True, 'b': {'c': {'e': True, 'd': [{'g': False}]}}}
        immutable_dict = ImmutableCaseInsensitiveDict(initial_dict)
        dict_copy = immutable_dict.copy()
        self.assertIsNot(dict_copy, initial_dict)
        self.assertIsNot(dict_copy['b'], initial_dict['b'])
        self.assertEqual(dict_copy['b'], initial_dict['b'])
        self.assertIsInstance(dict_copy['b'], dict)
        self.assertIsInstance(dict_copy['b']['c'], dict)
        self.assertIsInstance(dict_copy['b']['c']['d'], list)
        self.assertIsNot(dict_copy['b']['c']['d'], initial_dict['b']['c']['d'])  # type: ignore
        dict_copy['h'] = False
        self.assertIs(dict_copy['h'], False)


class TestImmutableList(TestCase):

    def setUp(self) -> None:
        self.initial_list = ['a', 'b', 'c']
        self.immutable_list = ImmutableList(self.initial_list)

    def test_raises_error_on_non_existent_index(self) -> None:
        with self.assertRaises(IndexError):
            _ = self.immutable_list[10]

    def test_assignment_not_allowed(self) -> None:
        with self.assertRaises(TypeError):
            # pylint: disable=E1137
            self.immutable_list[0] = 'd'  # type: ignore

    def test_getitem(self) -> None:
        self.assertEqual(self.immutable_list[0], self.initial_list[0])

    def test_equality(self) -> None:
        # List
        self.assertEqual(self.initial_list, self.immutable_list)
        # Tuple
        self.assertEqual(tuple(self.initial_list), self.immutable_list)
        # Same class
        self.assertEqual(ImmutableList(self.initial_list.copy()), self.immutable_list)

    def test_shallow_copy(self) -> None:
        self.assertEqual(list(self.immutable_list._container), self.initial_list)
        self.assertIsNot(self.immutable_list._container, self.initial_list)

    def test_ensure_immutable(self) -> None:
        initial_list = [[1, 2], [3, 4], {'a': {'b': 1}}]
        immutable_list = ImmutableList(initial_list)
        self.assertIsInstance(immutable_list[0], ImmutableList)
        self.assertIsInstance(immutable_list[2], ImmutableCaseInsensitiveDict)
        self.assertIsInstance(immutable_list[2]['a'], ImmutableCaseInsensitiveDict)

    def test_iteration_returns_immutable_objects(self) -> None:
        data = [{'b': {'c': True}}, {'d': {'c': False}}]
        immutable_list = ImmutableList(data)
        iterated_list = list(immutable_list)
        self.assertIsNot(iterated_list[0], data[0])
        self.assertIsNot(iterated_list[1], data[1])
        self.assertIsNot(iterated_list[0]['b'], data[0]['b'])
        self.assertIsInstance(iterated_list[0]['b'], ImmutableCaseInsensitiveDict)

    def test_copy(self) -> None:
        initial_list = [{'b': {'c': True}}, {'d': {'c': False}}]
        immutable_list = ImmutableList(initial_list)
        list_copy = immutable_list.copy()
        self.assertIsNot(list_copy, initial_list)
        self.assertIsNot(list_copy[0], initial_list[0])
        self.assertEqual(list_copy[0], initial_list[0])
        self.assertIsInstance(list_copy[0], dict)
        list_copy.append({'e': False})
        self.assertIs(list_copy[-1]['e'], False)


class TestImmutableNestedList(TestCase):

    def setUp(self) -> None:
        self.initial_dict = {'a': [1, 2]}
        self.immutable_dict = ImmutableCaseInsensitiveDict(self.initial_dict)

    def test_assignment_not_allowed(self) -> None:
        with self.assertRaises(TypeError):
            self.immutable_dict['a'][0] = 100

    def test_original_dict_not_mutated(self) -> None:
        _ = self.immutable_dict['a'][0]
        self.assertEqual(self.initial_dict, self.immutable_dict._container)

    def test_raises_error_for_non_existent_index(self) -> None:
        with self.assertRaises(IndexError):
            _ = self.immutable_dict['a'][2]


class TestJSONSerialization(TestCase):

    def test_immutable_list(self) -> None:
        initial_list = [1, 2, 3]
        immutable_list = ImmutableList(initial_list)
        self.assertEqual(json.dumps(initial_list), json.dumps(immutable_list, default=json_encoder))

    def test_immutable_dict(self) -> None:
        initial_dict = {'a': [1, 2, 3], 'b': {'c': True}}
        immutable_dict = ImmutableCaseInsensitiveDict(initial_dict)
        self.assertEqual(json.dumps(initial_dict), json.dumps(immutable_dict, default=json_encoder))

    def test_raises_type_error_for_nonserializable_object(self) -> None:
        with self.assertRaises(TypeError):
            json.dumps({'test_case': TestCase}, default=json_encoder)


class TestCaseInsensitiveLookup(TestCase):

    def setUp(self) -> None:
        self.data = {
            'Content-Encoding': 'gzip',
            'ACCEPT': '*/*',
            'X-Forwarded-For': '10.0.0.1, 10.0.0.2',
            'Request': {
                'HTTP_version': '1.1',
                'query': 'p=1&r=2'
            }
        }
        self.case_insensitive_dict = ImmutableCaseInsensitiveDict(self.data)

    def test_membership_check(self) -> None:
        self.assertIn('accept', self.case_insensitive_dict)
        self.assertIn('Accept', self.case_insensitive_dict)
        self.assertIn('ACCEPT', self.case_insensitive_dict)
        self.assertIn('CONTENT-Encoding', self.case_insensitive_dict)
        self.assertIn('REQUEST', self.case_insensitive_dict)
        self.assertNotIn('Content-Length', self.case_insensitive_dict)
        self.assertIn('http_version', self.case_insensitive_dict['request'])

    def test_immutable_return_value(self) -> None:
        self.assertIsInstance(self.case_insensitive_dict['request'], ImmutableCaseInsensitiveDict)
        self.assertEqual(self.case_insensitive_dict['request'], self.data['Request'])
        self.assertEqual(self.case_insensitive_dict['request']['Query'], 'p=1&r=2')

    def test_getitem(self) -> None:
        self.assertEqual(self.case_insensitive_dict['accept'], self.data['ACCEPT'])
        self.assertEqual(self.case_insensitive_dict['CONTENT-ENCODING'], self.data['Content-Encoding'])
        self.assertEqual(self.case_insensitive_dict['request']['http_version'], '1.1')
        with self.assertRaises(KeyError):
            _ = self.case_insensitive_dict['Unknown-key']

    def test_get(self) -> None:
        self.assertEqual(self.case_insensitive_dict.get('x-forwarded-for'), self.data['X-Forwarded-For'])
        self.assertEqual(self.case_insensitive_dict.get('X-FORWARDED-FOR'), self.data['X-Forwarded-For'])
        self.assertIsNone(self.case_insensitive_dict.get('Unknown-key'))
        self.assertEqual(self.case_insensitive_dict.get('Unknown-key', 1), 1)

    def test_original_keys_on_iteration(self) -> None:
        self.assertListEqual(list(self.data), list(self.case_insensitive_dict))
        self.assertListEqual(list(self.data.keys()), list(self.case_insensitive_dict.keys()))
        self.assertListEqual(list(self.data.items()), list(self.case_insensitive_dict.items()))

    def test_on_conflict_first_occurrence_returned(self) -> None:
        data = {'X-Forwarded-For': '10.0.0.1, 10.0.0.2', 'X-FORWARDED-FOR': '10.0.0.3, 10.0.0.4'}
        case_insensitive_dict = ImmutableCaseInsensitiveDict(data)
        self.assertEqual(case_insensitive_dict['X-FORWARDED-FOR'], '10.0.0.3, 10.0.0.4')
        self.assertEqual(case_insensitive_dict['x-forwarded-for'], data['X-Forwarded-For'])

    def test_on_empty_container_raises_key_error(self) -> None:
        case_insensitive_dict = ImmutableCaseInsensitiveDict({})
        with self.assertRaisesRegex(KeyError, 'non_existing_key'):
            _ = case_insensitive_dict['non_existing_key']

    def test_incremental_map_build(self) -> None:
        # Normally protected attributes should not participate in testing,
        # however this is critical functionality, so we maximize the test coverage.
        self.assertIn('CONTENT-ENCODING', self.case_insensitive_dict)
        self.assertEqual(self.case_insensitive_dict._case_insensitive_keymap, {'content-encoding': 'Content-Encoding'})
        self.assertIn('X-FORWARDED-FOR', self.case_insensitive_dict)
        self.assertEqual(
            self.case_insensitive_dict._case_insensitive_keymap, {
                'content-encoding': 'Content-Encoding',
                'accept': 'ACCEPT',
                'x-forwarded-for': 'X-Forwarded-For'
            }
        )
        # Ensure existing keys in the key map are returned
        self.assertIn('CONTENT-ENCODING', self.case_insensitive_dict)
        self.assertIn('REQUEST', self.case_insensitive_dict)
        self.assertEqual(
            self.case_insensitive_dict._case_insensitive_keymap, {
                'content-encoding': 'Content-Encoding',
                'accept': 'ACCEPT',
                'x-forwarded-for': 'X-Forwarded-For',
                'request': 'Request'
            }
        )
        # Ensure after the keymap is fully built keys are successfully matched
        self.assertIn('X-FORWARDED-FOR', self.case_insensitive_dict)
        self.assertTrue(self.case_insensitive_dict._keymap_fully_built)



import base64
import dataclasses
import enum
import json
from typing import Any, Dict, Final, List, Optional, Union

import panther_core.data_model
import panther_core.policy
import panther_core.rule

from panther_analysis_tool import util as pat_utils


class SdkContentType(enum.Enum):
    UNKNOWN = 1
    RULE = 2
    POLICY = 3
    SCHEDULED_RULE = 4
    DATA_MODEL = 5
    QUERY = 6


class PythonFunc:
    _PATH_TO_SRC: Final = ["src"]
    _PATH_TO_NAME: Final = ["name"]

    def __init__(self, _func: Dict):
        self.src = pat_utils.deep_get(_func, self._PATH_TO_SRC, "No source found")
        self.name = pat_utils.deep_get(_func, self._PATH_TO_NAME, "No name found")

    def get_code(self) -> str:
        return base64.standard_b64decode(self.src).decode("utf8")

    def get_name(self) -> str:
        return self.name


class DataModelMapping:
    _PATH_TO_NAME = ["d", "name"]
    _PATH_TO_PATH = ["d", "path"]
    _PATH_TO_FUNC = ["d", "func"]

    def __init__(self, _mapping: Dict):
        self.name: str = pat_utils.deep_get(_mapping, self._PATH_TO_NAME, "no name found")
        self.path: str = pat_utils.deep_get(_mapping, self._PATH_TO_PATH) or ""
        _func = pat_utils.deep_get(_mapping, self._PATH_TO_FUNC) or {}
        if not self.using_path():
            self.func = PythonFunc(_func)

    def using_path(self) -> bool:
        return self.path != ""

    def get_name(self, lowercase: bool = False) -> str:
        if lowercase:
            return self.name.lower()
        return self.name

    def get_func_name(self) -> Optional[str]:
        return self.func.get_name() if not self.using_path() else None

    def to_panther_core_mapping(self) -> Dict[str, Any]:
        return {
            "name": self.get_name(lowercase=True),
            "path": self.path,
            "method": self.get_func_name(),
        }


class DataModel:
    _PATH_TO_LOG_TYPE = ["val", "d", "log_type"]
    _PATH_TO_MAPPINGS = ["val", "d", "mappings"]
    _PATH_TO_ID = ["val", "d", "data_model_id"]
    _PATH_TO_ENABLED = ["val", "d", "enabled"]
    _PATH_TO_NAME = ["val", "d", "name"]

    def __init__(self, _data_model: Dict):
        self.log_type = pat_utils.deep_get(_data_model, self._PATH_TO_LOG_TYPE, "no log type found")
        _mappings = pat_utils.deep_get(_data_model, self._PATH_TO_MAPPINGS, [])
        self.mappings = [DataModelMapping(m) for m in _mappings]
        self.name = pat_utils.deep_get(_data_model, self._PATH_TO_NAME, "no name found")
        self.data_model_id = pat_utils.deep_get(_data_model, self._PATH_TO_ID, "no id found")
        self.enabled = bool(pat_utils.deep_get(_data_model, self._PATH_TO_ENABLED, False))

    def module_body(self) -> str:
        return "\n\n\n".join([m.func.get_code() for m in self.mappings if not m.using_path()])

    def to_panther_core_data_model(self) -> panther_core.data_model.DataModel:
        return panther_core.data_model.DataModel(
            {
                "id": self.data_model_id,
                "versionId": "",
                "mappings": [m.to_panther_core_mapping() for m in self.mappings],
                "body": self.module_body(),
            }
        )


class Filter:
    _PATH_TO_FILTER_FUNC: Final = ["d", "func"]

    def __init__(self, _filter: Dict):
        self.func = PythonFunc(pat_utils.deep_get(_filter, self._PATH_TO_FILTER_FUNC, {}))

    def get_code(self) -> str:
        return self.func.get_code()

    def get_name(self) -> str:
        return self.func.get_name()


class UnitTest:
    def __init__(self, test: Dict):
        self.origin: str = pat_utils.deep_get(test, ["o", "name"], "No origin found")
        self.data: Dict = json.loads(pat_utils.deep_get(test, ["d", "data"], "{}"))
        self.name: str = pat_utils.deep_get(test, ["d", "name"], "No name found")
        self.expect_match: bool = bool(pat_utils.deep_get(test, ["d", "expect_match"], True))
        self.fail_reasons: List[str] = []  # only used if test failed

    def add_fail_reason(self, reason: str) -> None:
        self.fail_reasons.append(reason)


class Detection:
    _PATH_TO_UNIT_TESTS: Final = ["val", "d", "unit_tests"]
    _PATH_TO_FILTERS: Final = ["val", "d", "filters"]
    _PATH_TO_RULE_ID: Final = ["val", "d", "rule_id"]
    _PATH_TO_POLICY_ID: Final = ["val", "d", "policy_id"]
    _PATH_TO_ENABLED: Final = ["val", "d", "enabled"]
    _PATH_TO_ORIGIN: Final = ["val", "o", "name"]

    def __init__(self, detection: Dict) -> None:
        self.detection_type = panther_sdk_key_to_type(detection.get("key") or "")

        tests = pat_utils.deep_get(detection, self._PATH_TO_UNIT_TESTS, [])
        self.unit_tests: List[UnitTest] = [UnitTest(t) for t in pat_utils.to_list(tests)]

        filts = pat_utils.deep_get(detection, self._PATH_TO_FILTERS, [])
        self.filters: List[Filter] = [Filter(f) for f in pat_utils.to_list(filts)]

        self.detection_id = pat_utils.deep_get(
            detection,
            self._PATH_TO_RULE_ID,
            pat_utils.deep_get(detection, self._PATH_TO_POLICY_ID, "No ID found"),
        )

        self.enabled = pat_utils.deep_get(detection, self._PATH_TO_ENABLED, False)

        self.origin = pat_utils.deep_get(detection, self._PATH_TO_ORIGIN, "No origin found")

    def has_unit_tests(self) -> bool:
        return len(self.unit_tests) > 0

    def has_pass_and_fail_tests(self) -> bool:
        pass_test, fail_test = False, False
        for unit_test in self.unit_tests:
            if unit_test.expect_match:
                pass_test = True
            if not unit_test.expect_match:
                fail_test = True
        return pass_test and fail_test

    def disabled(self) -> bool:
        return not self.enabled

    def module_body(self) -> str:
        prg = "_filters = [] \n\n"
        for filt in self.filters:
            prg += f"{filt.get_code()} \n\n"
            prg += f"_filters.append({filt.get_name()}) \n\n"

        match_val = False  # flip the return vals if it is a policy
        matcher_func_name = "rule"
        if self.detection_type is SdkContentType.POLICY:
            match_val = True
            matcher_func_name = "policy"

        prg += f"def {matcher_func_name}(event):\n"
        prg += "    global _filters\n"
        prg += "    for f in _filters:\n"
        prg += f"        if f(event) == {match_val}:\n"
        prg += f"            return {match_val}\n"
        prg += f"    return {not match_val}\n"

        return prg

    def to_panther_core_config(self) -> Dict[str, Any]:
        return {
            "id": self.detection_id,
            "versionId": "",
            "body": self.module_body(),
            "analysisType": self.detection_type.name,
        }

    def to_panther_core_detection(
        self,
    ) -> Union[panther_core.rule.Rule, panther_core.policy.Policy]:
        if self.detection_type is SdkContentType.POLICY:
            return panther_core.policy.Policy(self.to_panther_core_config())
        return panther_core.rule.Rule(self.to_panther_core_config())


@dataclasses.dataclass
class SdkContent:
    detections: List[Detection] = dataclasses.field(default_factory=list)
    data_models: List[DataModel] = dataclasses.field(default_factory=list)


def panther_sdk_key_to_type(key: str) -> SdkContentType:
    if "rule" in key:
        return SdkContentType.RULE
    if "policy" in key:
        return SdkContentType.POLICY
    if "scheduled-rule" in key:
        return SdkContentType.SCHEDULED_RULE
    if "data-model" in key:
        return SdkContentType.DATA_MODEL
    if "query" in key:
        return SdkContentType.QUERY
    return SdkContentType.UNKNOWN



import json
import logging
import os
from fnmatch import fnmatch
from typing import Any, Dict, Iterator, List, Tuple

from ruamel.yaml import YAML
from ruamel.yaml import parser as YAMLParser
from ruamel.yaml import scanner as YAMLScanner

from panther_analysis_tool.constants import (
    DATA_MODEL_PATH_PATTERN,
    HELPERS_PATH_PATTERN,
    LUTS_PATH_PATTERN,
    PACKS_PATH_PATTERN,
    POLICIES_PATH_PATTERN,
    QUERIES_PATH_PATTERN,
    RULES_PATH_PATTERN,
)


def filter_analysis(
    analysis: List[Any], filters: Dict[str, List], filters_inverted: Dict[str, List]
) -> List[Any]:
    if filters is None:
        return analysis

    filtered_analysis = []
    for file_name, dir_name, analysis_spec in analysis:
        if fnmatch(dir_name, HELPERS_PATH_PATTERN):
            logging.debug("auto-adding helpers file %s", os.path.join(file_name))
            filtered_analysis.append((file_name, dir_name, analysis_spec))
            continue
        if fnmatch(dir_name, DATA_MODEL_PATH_PATTERN):
            logging.debug("auto-adding data model file %s", os.path.join(file_name))
            filtered_analysis.append((file_name, dir_name, analysis_spec))
            continue
        match = True
        for key, values in filters.items():
            spec_value = analysis_spec.get(key, "")
            spec_value = spec_value if isinstance(spec_value, list) else [spec_value]
            if not set(spec_value).intersection(values):
                match = False
                break
        for key, values in filters_inverted.items():
            spec_value = analysis_spec.get(key, "")
            spec_value = spec_value if isinstance(spec_value, list) else [spec_value]
            if set(spec_value).intersection(values):
                match = False
                break

        if match:
            filtered_analysis.append((file_name, dir_name, analysis_spec))

    return filtered_analysis


def load_analysis_specs(
    directories: List[str], ignore_files: List[str]
) -> Iterator[Tuple[str, str, Any, Any]]:
    """Loads the analysis specifications from a file.

    Args:
        directories: The relative path to Panther policies or rules.
        ignore_files: Files that Panther Analysis Tool should not process

    Yields:
        A tuple of the relative filepath, directory name, and loaded analysis specification dict.
    """
    # setup a list of paths to ensure we do not import the same files
    # multiple times, which can happen when testing from root directory without filters
    ignored_normalized = []
    for file in ignore_files:
        ignored_normalized.append(os.path.normpath(file))

    loaded_specs: List[Any] = []
    for directory in directories:
        for relative_path, _, file_list in os.walk(directory):
            # Skip hidden folders
            if (
                relative_path.split("/")[-1].startswith(".")
                and relative_path != "./"
                and relative_path != "."
            ):
                continue
            # setup yaml object
            yaml = YAML(typ="safe")
            # If the user runs with no path args, filter to make sure
            # we only run folders with valid analysis files. Ensure we test
            # files in the current directory by not skipping this iteration
            # when relative_path is the current dir
            if directory in [".", "./"] and relative_path not in [".", "./"]:
                if not any(
                    (
                        fnmatch(relative_path, path_pattern)
                        for path_pattern in (
                            DATA_MODEL_PATH_PATTERN,
                            HELPERS_PATH_PATTERN,
                            LUTS_PATH_PATTERN,
                            RULES_PATH_PATTERN,
                            PACKS_PATH_PATTERN,
                            POLICIES_PATH_PATTERN,
                            QUERIES_PATH_PATTERN,
                        )
                    )
                ):
                    logging.debug("Skipping path %s", relative_path)
                    continue
            for filename in sorted(file_list):
                # Skip hidden files
                if filename.startswith("."):
                    continue
                spec_filename = os.path.abspath(os.path.join(relative_path, filename))
                # skip loading files that have already been imported
                if spec_filename in loaded_specs:
                    continue
                # Dont load files that are explictly ignored
                relative_name = os.path.normpath(os.path.join(relative_path, filename))
                if relative_name in ignored_normalized:
                    logging.info("ignoring file %s", relative_name)
                    continue
                loaded_specs.append(spec_filename)
                if fnmatch(filename, "*.y*ml"):
                    with open(spec_filename, "r") as spec_file_obj:
                        try:
                            yield spec_filename, relative_path, yaml.load(spec_file_obj), None
                        except (YAMLParser.ParserError, YAMLScanner.ScannerError) as err:
                            # recreate the yaml object and yield the error
                            yaml = YAML(typ="safe")
                            yield spec_filename, relative_path, None, err
                if fnmatch(filename, "*.json"):
                    with open(spec_filename, "r") as spec_file_obj:
                        try:
                            yield spec_filename, relative_path, json.load(spec_file_obj), None
                        except ValueError as err:
                            yield spec_filename, relative_path, None, err


def to_relative_path(filename: str) -> str:
    cwd = os.getcwd()
    return os.path.relpath(filename, cwd)
